#!/usr/bin/python

import os
import time
from ConfigParser import ConfigParser
import subprocess
from socket import gethostname

from fjd.core_process import CoreProcess
from fjd.utils import ensure_wdir


class Worker(CoreProcess):
    '''
    A worker process
    '''

    def __init__(self, interval=.1, project=None):
        if not project:
            project = 'default'
        self.wdir = ensure_wdir(project)
        self.start_up()

        # announce my presence
        self.id = self.mk_id()
        print('[fjd-worker] Started with ID {id}.'.format(id=self.id))
        subprocess.call('touch {wdir}/workerqueue/{id}.worker'\
                   .format(wdir=self.wdir, id=self.id), shell=True)

        # look for jobs
        while True:
            job = self.next_job_on_pod()
            if job:
                print('[fjd-worker] Worker {}: I found a job.'.format(self.id))
                # A job is a config file
                conf = ConfigParser() 
                conf.read('{}/jobpod/{}'.format(self.wdir, job))
                exe = conf.get('control', 'executable')
                exe_log = conf.get('control', 'logfile')
                # make log file and execute task
                cmd = 'touch {log}; {exe} {wdir}/jobpod/{job}; '\
                       .format(log=exe_log, exe=exe, wdir=self.wdir, job=job)
                subprocess.call(cmd, shell=True)
                print('[fjd-worker] Worker {}: Finished my job.'.format(self.id))
                # remove the job from pod (signaling it is done) + re-announce myself
                subprocess.call('rm {wdir}/jobpod/{job}; touch {wdir}/workerqueue/{id}.worker'\
                        .format(wdir=self.wdir, job=job, id=self.id), shell=True)
            time.sleep(interval)

    def mk_id(self):
        ms = time.time()
        return '{h}_{t}'.format(h=gethostname(), t=ms)

    def next_job_on_pod(self):
        my_jobs = [j for j in os.listdir('{}/jobpod'.format(self.wdir))\
                  if j.startswith(self.id)]
        if len(my_jobs) > 0:
            return my_jobs[0]
        else:
            return None




#!/usr/bin/env python

import sys
import os
import time
import signal

from fjd import CoreProcess
from fjd import Recruiter
from fjd.utils import ensure_wdir


class Dispatcher(CoreProcess):
    '''
    Directory and File-based work dispatcher.
    Expects jobs to be put in directory `jobqueue` and workers to
    annouce themselves in directory `workerqueue`. Will then assign jobs 
    to workers by moving jobs to the `jobpod` directory. 
    '''

    def __init__(self, interval=.1, project=None, end_on_empty_queue=True):
        if not project:
            project = 'default'
        self.wdir = ensure_wdir(project)
        self.start_up()

        print('[fjd-dispatcher] Started on project "{}"'.format(project))

        def signal_handler(signal, frame):
            ''' gently exiting, e.g. when CTRL-C was pressed.  '''
            sys.stdout.write('\n[fjd-dispatcher] Received Exit signal. Bye.\n')
            sys.exit(0)
        signal.signal(signal.SIGINT, signal_handler)

        do_work = True
        while do_work:
            time.sleep(interval)
            jq = os.listdir('{}/jobqueue'.format(self.wdir))
            jp = os.listdir('{}/jobpod'.format(self.wdir))
            wq = os.listdir('{}/workerqueue'.format(self.wdir))
            if len(jq) > 0:  # more jobs waiting for workers
                sys.stdout.write("\r[fjd-dispatcher] {} job(s) waiting in the queue. Currently {} worker(s) are free...  "\
                       .format(len(jq), len(wq)))
                sys.stdout.flush()
                for _ in range(min(len(jq), len(wq))):
                    worker = wq.pop()
                    job = jq.pop()
                    os.rename('{wdir}/jobqueue/{j}'.format(wdir=self.wdir, j=job),
                              '{wdir}/jobpod/{w}'.format(wdir=self.wdir, w=worker))
                    os.remove('{wdir}/workerqueue/{w}'.format(wdir=self.wdir, w=worker))
            elif len(jp) > 0:  # some jobs are still running
                sys.stdout.write("\r[fjd-dispatcher] Queue is empty. Waiting for remaining {} job(s) to finish ...        ".format(len(jp)))
                sys.stdout.flush()
            else:  # all jobs are done
                sys.stdout.write("\r[fjd-dispatcher] Queue is empty and all jobs have finished.                          ")
                sys.stdout.flush()
                if end_on_empty_queue:
                    Recruiter(project=project).fire()
                    do_work = False
        self.wrap_up()




from getpass import getpass
import time
# ignore deprecation warnings that paramiko currently delivers
import warnings
warnings.simplefilter("ignore", DeprecationWarning)
import paramiko


'''
Tools for SSH communication: make client, send commands
'''

def ssh(client, cmd, ignore=[]):
    '''
    Run cmd on remote client (retry if connection fails), show errors (if we care about them)

    :params paramiko.SSHClient client:
    :params string cmd:
    :params string ignore: Error messages to ignore for this cmd
    :returns: stdout from host (minus some things we consider irrelevant)
    '''
    done = False
    while not done:
        try:
            stdin, stdout, stderr = client.exec_command(cmd)
            done = True
        except:
            time.sleep(4)
    err = stderr.read()
    dontcare_snippets = ['xset:', 'cannot remove']
    dontcare_snippets.extend(ignore)
    err_out = ""
    for e_line in err.split('\n'):
        yell_it = True
        for s in dontcare_snippets:
            if s in e_line:
                yell_it = False
        if yell_it and e_line != "":
            err_out += '%s\n' % e_line
    if err_out.strip() != "":
        print("[fjd-ssh] Error while doing stuff on server: {}".format(err_out))
    return stdout.read()


def mk_ssh_client(hostname, username):
    '''
    Make an SSH client and connect it.
    We first try a passwordless login, and then ask for credentials.

    :param string hostname: name of host
    :param string username: username on host
    :returns: paramiko.SSHClient if successful, None otherwise
    '''
    ssh_client = paramiko.SSHClient()
    ssh_client.load_system_host_keys()
    ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        ssh_client.connect(hostname, username=username)
    except (paramiko.AuthenticationException, paramiko.SSHException):
        print "[fjd-ssh] Could not connect to host '%s' as user '%s' with no password." % (hostname, username)
        print "          If you want password-less logon, please check your RSA key or shared/remembered connection setup."
    except Exception, e:
        print("[fjd-ssh] WARNING: Error while connecting with host {}: {}. ".format(hostname, e))
        if "Unknown server" in str(e):
            print "          Is it a known host (look in ~/.ssh/known_hosts)?"
        return None
    else:
        return ssh_client
    ssh_client = None
    while ssh_client is None:
        print("[fjd-ssh] Logging in user '{}' on host '{}' now (type 'exit' to abort): ".format(username, hostname))
        passwd = getpass()
        if passwd == 'exit':
            break
        ssh_client = paramiko.SSHClient()
        ssh_client.load_system_host_keys()
        try:
            ssh_client.connect(hostname, username=username, password=passwd)
        except paramiko.AuthenticationException:
            print("[fjd-ssh] Authentication was not successful.")
            ssh_client = None
        except Exception, e:
            print("[fjd-ssh] WARNING: Error while connecting with host {}: {}".format((hostname, e)))
            ssh_client = None
    return ssh_client






class CoreProcess(object):
    '''
    Superclass of classes that represent a longer-running process, i.e.
    Worker and Dispatcher.

    Not sure if needed - we could do things here that are needed for all
    processes on a cluster core, e.g. notify cluster management that we exist 
    or that we stopped.
    '''

    def start_up(self):
        pass

    def wrap_up(self):
        pass
